# Copyright 2021 ONDEWO GmbH
#
# Licensed under the Apache License, Version 2.0 (the License);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an AS IS BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
import os

TIMEOUT_MINUTES: int = int(os.getenv("TIMEOUT_MINUTES", "20"))
PROVIDER_STT: Optional[str] = os.getenv("SIP_PROVIDER_STT")
PROVIDER_TTS: Optional[str] = os.getenv("SIP_PROVIDER_TTS")
LANG: Optional[str] = os.getenv("SIP_LANG")
